@@ -211,11 +211,14 @@ struct _GarconMenuPrivate
   GarconMenuItemCache *cache;
 
   /* List to merge consecutive file changes into a a single event */
-  GList               *changed_files;
+  GSList              *changed_files;
   guint                file_changed_idle;
 
   /* Flag for marking custom path menus */
   guint                uses_custom_path : 1;
+
+  /* idle reload-required to group events */
+  guint                idle_reload_required_id;
 };
 
 
@@ -330,6 +333,7 @@ garcon_menu_init (GarconMenu *menu)
   menu->priv->pool = garcon_menu_item_pool_new ();
   menu->priv->uses_custom_path = TRUE;
   menu->priv->changed_files = NULL;
+  menu->priv->idle_reload_required_id = 0;
 
   /* Take reference on the menu item cache */
   menu->priv->cache = garcon_menu_item_cache_get_default ();
@@ -353,19 +357,16 @@ garcon_menu_clear (GarconMenu *menu)
       menu->priv->tree = NULL;
 
       /* Release the merge files */
-      g_list_foreach (menu->priv->merge_files, (GFunc) g_object_unref, NULL);
-      g_list_free (menu->priv->merge_files);
+      _garcon_g_list_free_full (menu->priv->merge_files, g_object_unref);
       menu->priv->merge_files = NULL;
 
       /* Release the merge dirs */
-      g_list_foreach (menu->priv->merge_dirs, (GFunc) g_object_unref, NULL);
-      g_list_free (menu->priv->merge_dirs);
+      _garcon_g_list_free_full (menu->priv->merge_dirs, g_object_unref);
       menu->priv->merge_dirs = NULL;
     }
 
   /* Free submenus */
-  g_list_foreach (menu->priv->submenus, (GFunc) g_object_unref, NULL);
-  g_list_free (menu->priv->submenus);
+  _garcon_g_list_free_full (menu->priv->submenus, g_object_unref);
   menu->priv->submenus = NULL;
 
   /* Free directory */
@@ -377,6 +378,13 @@ garcon_menu_clear (GarconMenu *menu)
 
   /* Clear the item pool */
   garcon_menu_item_pool_clear (menu->priv->pool);
+
+  /* Stop reload-required emit */
+  if (menu->priv->idle_reload_required_id != 0)
+    {
+      g_source_remove (menu->priv->idle_reload_required_id);
+      menu->priv->idle_reload_required_id = 0;
+    }
 }
 
 
@@ -1177,8 +1185,8 @@ garcon_menu_collect_files_from_path (GarconMenu  *menu,
 
 
 static gboolean
-collect_rules (GNode  *node,
-               GList **list)
+collect_rules (GNode   *node,
+               GSList **list)
 {
   GarconMenuNodeType type;
 
@@ -1187,7 +1195,7 @@ collect_rules (GNode  *node,
   if (type == GARCON_MENU_NODE_TYPE_INCLUDE ||
       type == GARCON_MENU_NODE_TYPE_EXCLUDE)
     {
-      *list = g_list_append (*list, node);
+      *list = g_slist_append (*list, node);
     }
 
   return FALSE;
@@ -1200,14 +1208,16 @@ garcon_menu_resolve_items (GarconMenu *menu,
                            GHashTable *desktop_id_table,
                            gboolean    only_unallocated)
 {
-  GList  *rules = NULL;
-  GList  *iter;
+  GSList  *rules = NULL;
+  GSList  *iter;
+  GList   *submenu;
   gboolean menu_only_unallocated = FALSE;
 
-  g_return_if_fail (menu != NULL && GARCON_IS_MENU (menu));
+  g_return_if_fail (GARCON_IS_MENU (menu));
 
-  menu_only_unallocated = garcon_menu_node_tree_get_boolean_child (menu->priv->tree,
-                                                                   GARCON_MENU_NODE_TYPE_ONLY_UNALLOCATED);
+  menu_only_unallocated = 
+    garcon_menu_node_tree_get_boolean_child (menu->priv->tree,
+                                             GARCON_MENU_NODE_TYPE_ONLY_UNALLOCATED);
 
   /* Resolve items in this menu (if it matches the only_unallocated argument.
    * This means that in the first pass, all items of menus without
@@ -1219,7 +1229,7 @@ garcon_menu_resolve_items (GarconMenu *menu,
                        (GNodeTraverseFunc) collect_rules, &rules);
 
       /* Iterate over all rules */
-      for (iter = rules; iter != NULL; iter = g_list_next (iter))
+      for (iter = rules; iter != NULL; iter = g_slist_next (iter))
         {
           if (G_LIKELY (garcon_menu_node_tree_get_node_type (iter->data) == GARCON_MENU_NODE_TYPE_INCLUDE))
             {
@@ -1232,13 +1242,17 @@ garcon_menu_resolve_items (GarconMenu *menu,
               garcon_menu_item_pool_apply_exclude_rule (menu->priv->pool, iter->data);
             }
         }
+
+      /* Cleanup */
+      g_slist_free (rules);
     }
 
   /* Iterate over all submenus */
-  for (iter = menu->priv->submenus; iter != NULL; iter = g_list_next (iter))
+  for (submenu = menu->priv->submenus; submenu != NULL; submenu = g_list_next (submenu))
     {
       /* Resolve items of the submenu */
-      garcon_menu_resolve_items (GARCON_MENU (iter->data), desktop_id_table, only_unallocated);
+      garcon_menu_resolve_items (GARCON_MENU (submenu->data), desktop_id_table,
+                                 only_unallocated);
     }
 }
 
@@ -1488,12 +1502,12 @@ layout_elements_collect (GList **dest_list,
       if (GARCON_IS_MENU (iter->data))
         {
           if (G_LIKELY (!layout_has_menuname (layout, garcon_menu_get_name (iter->data))))
-            *dest_list = g_list_append (*dest_list, iter->data);
+            *dest_list = g_list_prepend (*dest_list, iter->data);
         }
       else if (GARCON_IS_MENU_ITEM (iter->data))
         {
           if (G_LIKELY (!layout_has_filename (layout, garcon_menu_item_get_desktop_id (iter->data))))
-            *dest_list = g_list_append (*dest_list, iter->data);
+            *dest_list = g_list_prepend (*dest_list, iter->data);
         }
     }
 }
@@ -1544,7 +1558,7 @@ garcon_menu_get_elements (GarconMenu *menu)
 
           /* If the item with this desktop ID is included in the menu, append it to the list */
           if (G_LIKELY (item != NULL))
-            items = g_list_append (items, item);
+            items = g_list_prepend (items, item);
         }
       else if (type == GARCON_MENU_NODE_TYPE_MENUNAME)
         {
@@ -1554,12 +1568,12 @@ garcon_menu_get_elements (GarconMenu *menu)
 
           /* If there is such a menu, append it to the list */
           if (G_LIKELY (submenu != NULL))
-            items = g_list_append (items, submenu);
+            items = g_list_prepend (items, submenu);
         }
       else if (type == GARCON_MENU_NODE_TYPE_SEPARATOR)
         {
           /* Append separator to the list */
-          items = g_list_append (items, garcon_menu_separator_get_default ());
+          items = g_list_prepend (items, garcon_menu_separator_get_default ());
         }
       else if (type == GARCON_MENU_NODE_TYPE_MERGE)
         {
@@ -1568,38 +1582,47 @@ garcon_menu_get_elements (GarconMenu *menu)
 
           if (merge_type == GARCON_MENU_LAYOUT_MERGE_ALL)
             {
-              /* Get all menu items of this menu */
-              menu_items = garcon_menu_get_items (menu);
+              /* Get all the submenus */
+              menu_items = g_list_copy (menu->priv->submenus);
 
-              /* Append submenus */
-              menu_items = g_list_concat (menu_items, garcon_menu_get_menus (menu));
+              /* Get all menu items of this menu */
+              garcon_menu_item_pool_foreach (menu->priv->pool, (GHFunc) items_collect, &menu_items);
 
               /* Sort menu items */
               menu_items = g_list_sort (menu_items, (GCompareFunc) garcon_menu_compare_items);
 
-              /* Append menu items to the returned item list */
+              /* Prepend menu items to the returned item list */
               layout_elements_collect (&items, menu_items, layout);
+
+              /* Cleanup */
+              g_list_free (menu_items);
             }
           else if (merge_type == GARCON_MENU_LAYOUT_MERGE_FILES)
             {
               /* Get all menu items of this menu */
               menu_items = garcon_menu_get_items (menu);
 
-              /* Append menu items to the returned item list */
+              /* Prepend menu items to the returned item list */
               layout_elements_collect (&items, menu_items, layout);
+
+              /* Cleanup */
+              g_list_free (menu_items);
             }
           else if (merge_type == GARCON_MENU_LAYOUT_MERGE_MENUS)
             {
               /* Get all submenus */
               menu_items = garcon_menu_get_menus (menu);
 
-              /* Append submenus to the returned item list */
+              /* Prepend submenus to the returned item list */
               layout_elements_collect (&items, menu_items, layout);
+
+              /* Cleanup */
+              g_list_free (menu_items);
             }
         }
     }
 
-  return items;
+  return g_list_reverse (items);
 }
 
 
@@ -1814,8 +1837,7 @@ garcon_menu_stop_monitoring (GarconMenu *menu)
     g_source_remove (menu->priv->file_changed_idle);
 
   /* Free the hash table for merging consecutive file change events */
-  g_list_foreach (menu->priv->changed_files, (GFunc) g_object_unref, NULL);
-  g_list_free (menu->priv->changed_files);
+  _garcon_g_slist_free_full (menu->priv->changed_files, g_object_unref);
   menu->priv->changed_files = NULL;
 }
 
@@ -1960,8 +1982,7 @@ garcon_menu_monitor_app_dirs (GarconMenu *menu)
   garcon_menu_monitor_files (menu, dirs, garcon_menu_app_dir_changed);
 
   /* Release the allocated GFiles and free the list */
-  g_list_foreach (dirs, (GFunc) g_object_unref, NULL);
-  g_list_free (dirs);
+  _garcon_g_list_free_full (dirs, g_object_unref);
 
   /* Free app dir list */
   g_list_free (app_dirs);
@@ -2030,6 +2051,34 @@ garcon_menu_monitor_directory_dirs (GarconMenu *menu)
 
 
 
+static gboolean
+garcon_menu_file_emit_reload_required_idle (gpointer data)
+{
+  GarconMenu *menu = GARCON_MENU (data);
+
+  g_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
+
+  menu->priv->idle_reload_required_id = 0;
+
+  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+
+  return FALSE;
+}
+
+
+
+static void
+garcon_menu_file_emit_reload_required (GarconMenu *menu)
+{
+  if (menu->priv->idle_reload_required_id == 0)
+    {
+      menu->priv->idle_reload_required_id = 
+        g_idle_add (garcon_menu_file_emit_reload_required_idle, menu);
+    }
+}
+
+
+
 static void
 garcon_menu_file_changed (GarconMenu       *menu,
                           GFile            *file,
@@ -2051,7 +2100,7 @@ garcon_menu_file_changed (GarconMenu       *menu,
   /* Quick check: reloading is needed if the menu file being used has changed */
   if (g_file_equal (menu->priv->file, file))
     {
-      g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+      garcon_menu_file_emit_reload_required (menu);
       return;
     }
 
@@ -2065,7 +2114,9 @@ garcon_menu_file_changed (GarconMenu       *menu,
   paths = garcon_config_build_paths (relative_filename);
 
   /* Check if the event file has higher priority than the file currently being used */
-  for (n = 0; !lower_priority && !higher_priority && paths != NULL && paths[n] != NULL; ++n) 
+  for (n = 0; 
+       !lower_priority && !higher_priority && paths != NULL && paths[n] != NULL;
+       ++n) 
     {
       menu_file = g_file_new_for_path (paths[n]);
 
@@ -2089,8 +2140,9 @@ garcon_menu_file_changed (GarconMenu       *menu,
   g_free (relative_filename);
 
   /* If the event file has higher priority, a menu reload is needed */
-  if (!lower_priority && higher_priority)
-    g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+  if (!lower_priority
+      && higher_priority)
+    garcon_menu_file_emit_reload_required (menu);
 }
 
 
@@ -2105,7 +2157,7 @@ garcon_menu_merge_file_changed (GarconMenu       *menu,
   g_return_if_fail (GARCON_IS_MENU (menu));
   g_return_if_fail (menu->priv->parent == NULL);
 
-  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+  garcon_menu_file_emit_reload_required (menu);
 }
 
 
@@ -2120,7 +2172,7 @@ garcon_menu_merge_dir_changed (GarconMenu       *menu,
   g_return_if_fail (GARCON_IS_MENU (menu));
   g_return_if_fail (menu->priv->parent == NULL);
 
-  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+  garcon_menu_file_emit_reload_required (menu);
 }
 
 
@@ -2133,7 +2185,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
   gboolean        affects_the_outside = FALSE;
   gboolean        stop_processing = FALSE;
   GFile          *file;
-  GList          *lp;
+  GSList         *lp;
   gchar          *path;
 
   g_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
@@ -2154,7 +2206,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
            * - created (possibly inside an existing one)
            * this is not trivial to handle, so we simply enforce a
            * menu reload to deal with the changes */
-          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+          garcon_menu_file_emit_reload_required (menu);
           stop_processing = TRUE;
         }
       else
@@ -2175,7 +2227,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
                            * moved around between different menus. this is slightly
                            * more complicated than one would first think, so just
                            * enforce a complete menu reload for now */
-                          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                          garcon_menu_file_emit_reload_required (menu);
                           stop_processing = TRUE;
                         }
                       else
@@ -2199,7 +2251,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
                        * or that the file was deleted. handling most situations can be very
                        * tricky, so, again, we just enfore a menu reload until we have 
                        * something better */
-                      g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                      garcon_menu_file_emit_reload_required (menu);
                       stop_processing = TRUE;
                     }
                 }
@@ -2208,7 +2260,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
                   /* there could be a lot of stuff happening here. seriously, this
                    * stuff is complicated. for now, simply enforce a complete reload 
                    * of the menu structure */
-                  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                  garcon_menu_file_emit_reload_required (menu);
                   stop_processing = TRUE;
                 }
             }
@@ -2217,8 +2269,7 @@ garcon_menu_process_file_changes (GarconMenu *menu)
     }
 
   /* reset the changed files list, all events processed */
-  g_list_foreach (menu->priv->changed_files, (GFunc) g_object_unref, NULL);
-  g_list_free (menu->priv->changed_files);
+  _garcon_g_slist_free_full (menu->priv->changed_files, g_object_unref);
   menu->priv->changed_files = NULL;
 
   /* reset the idle source ID */
@@ -2258,10 +2309,10 @@ garcon_menu_app_dir_changed (GarconMenu       *menu,
     {
       /* add the file to the changed files queue if we have no change event for
        * it queued yet */
-      if (g_list_find_custom (menu->priv->changed_files, file, compare_files) == NULL)
+      if (g_slist_find_custom (menu->priv->changed_files, file, compare_files) == NULL)
         {
-          menu->priv->changed_files = g_list_append (menu->priv->changed_files, 
-                                                     g_object_ref (file));
+          menu->priv->changed_files = g_slist_prepend (menu->priv->changed_files, 
+                                                       g_object_ref (file));
 
           /* register the idle handler if it is not active yet */
           if (menu->priv->file_changed_idle == 0)
@@ -2281,7 +2332,7 @@ garcon_menu_app_dir_changed (GarconMenu       *menu,
           /* an existing app dir (or a subdirectory) has been deleted. we
            * could remove all the items that are in use and reside inside
            * this root directory. but for now... enforce a menu reload! */
-          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+          garcon_menu_file_emit_reload_required (menu);
         }
       else 
         {
@@ -2297,7 +2348,7 @@ garcon_menu_app_dir_changed (GarconMenu       *menu,
                * yet whether there is a replacement in another app dir
                * with lower priority. we could try to find out but for now
                * it's easier to simply enforce a menu reload */
-              g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+              garcon_menu_file_emit_reload_required (menu);
             }
           else
             {
