@@ -2,7 +2,7 @@
 /*-
  * vi:set et ai sts=2 sw=2 cindent:
  *
- * Copyright (c) 2007 Jannis Pohlmann <jannis@xfce.org>
+ * Copyright (c) 2007-2011 Jannis Pohlmann <jannis@xfce.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -39,10 +39,12 @@ print_menu (GarconMenu  *menu,
             const gchar *path)
 {
   GarconMenuDirectory *directory;
+  GFile               *file;
   GList               *menus;
   GList               *items;
   GList               *iter;
   gchar               *name;
+  gchar               *file_path;
 
   if (!garcon_menu_element_get_visible (GARCON_MENU_ELEMENT (menu)))
     return;
@@ -83,8 +85,14 @@ print_menu (GarconMenu  *menu,
       if (GARCON_IS_MENU_ITEM (iter->data)
           && garcon_menu_element_get_visible (iter->data))
         {
+          file = garcon_menu_item_get_file (iter->data);
+          file_path = g_file_get_path (file);
+
           g_printf ("%s\t%s\t%s\n", name, garcon_menu_item_get_desktop_id (iter->data),
-                    garcon_menu_item_get_path (iter->data));
+                    file_path);
+
+          g_free (file_path);
+          g_object_unref (file);
         }
     }
 
